/* Copyright (C) 2007-2022 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Maksym Zyzak, Valentina Akishina, Igor Kulakov [committer], Sergei Zharko */

#pragma once  // include this header only once per compilation unit

#include "CaBranch.h"
#include "CaCloneMerger.h"
#include "CaConstants.h"
#include "CaField.h"
#include "CaGrid.h"
#include "CaGridEntry.h"
#include "CaHit.h"
#include "CaInputData.h"
#include "CaMeasurementXy.h"
#include "CaParameters.h"
#include "CaStation.h"
#include "CaTimer.h"
#include "CaTrack.h"
#include "CaTrackExtender.h"
#include "CaTrackFinder.h"
#include "CaTrackFinderWindow.h"
#include "CaTrackFitter.h"
#include "CaTrackParam.h"
#include "CaTrackingMonitor.h"
#include "CaTriplet.h"
#include "CaVector.h"
#include "CaWindowData.h"

#include <array>
#include <iomanip>
#include <iostream>
#include <limits>
#include <map>

namespace cbm::algo::ca
{
  class TripletConstructor;


  namespace
  {
    using namespace cbm::algo;  // to keep ca:: prefices in the code
  }

  //namespace
  //{
  //  using cbm::algo::ca::Track;      // TMP
  //  using cbm::algo::ca::Vector;     // TMP
  //  using cbm::algo::ca::Iteration;  // TMP
  //}  // namespace

  // *******************************
  // ** Types definition (global) **
  // *******************************

  using CaStationsArray_t = std::array<ca::Station<fvec>, constants::size::MaxNstations>;
  using CaMaterialArray_t = std::array<ca::MaterialMap, constants::size::MaxNstations>;
  using Tindex            = int;  // TODO: Replace with ca::HitIndex_t, if suitable

  struct CaHitTimeInfo {
    fscal fEventTimeMin{-std::numeric_limits<fscal>::max() / 2.};
    fscal fEventTimeMax{std::numeric_limits<fscal>::max() / 2.};
    fscal fMaxTimeBeforeHit{0.};  //< max event time for hits [0 .. hit] in the station hit array
    fscal fMinTimeAfterHit{0.};   //< min event time for hits [hit ... ] in the station hit array
  };

  /// Main class of CA track finder algorithm
  ///
  class Framework {
   public:
    // *************************
    // ** Friend classes list **
    // *************************

    friend class ca::TripletConstructor;
    friend class ca::TrackExtender;
    friend class ca::TrackFitter;
    friend class ca::TrackFinderWindow;
    friend class ca::TrackFinder;

    ///---------------------------
    /// Internal enumerations

    enum TrackingMode
    {
      kSts,
      kGlobal,
      kMcbm
    };

    // **********************************
    // ** Member functions declaration **
    // **********************************

    // ** Constructors and destructor

    /// Constructor
    Framework();

    /// Copy constructor
    Framework(const Framework&) = delete;

    /// Move constructor
    Framework(Framework&&) = delete;

    /// Copy assignment operator
    Framework& operator=(const Framework&) = delete;

    /// Move assignment operator
    Framework& operator=(Framework&&) = delete;

    /// Destructor
    ~Framework() = default;


    // ** Functions, which pack and unpack indexes of station and triplet **

    // TODO: move to ca::Triplet class (S.Zharko)
    /// Packs station and triplet indices to an unique triplet ID
    /// \param  iStation  Index of station in the active stations array
    /// \param  iTriplet  Index of triplet
    static unsigned int PackTripletId(unsigned int iStation, unsigned int iTriplet);

    /// Unpacks the triplet ID to its station index
    /// \param  id  Unique triplet ID
    static unsigned int TripletId2Station(unsigned int id);

    /// Unpacks the triplet ID to its triplet index
    /// \param  id  Unique triplet ID
    static unsigned int TripletId2Triplet(unsigned int id);


    /// Sets Framework parameters object
    /// \param other - reference to the Parameters object
    void SetParameters(const Parameters<fvec>& other) { fParameters = other; }
    // TODO: remove it (S.Zharko)

    /// Gets a pointer to the Framework parameters object
    const Parameters<fvec>& GetParameters() const { return fParameters; }

    /// Receives input data
    void ReceiveInputData(InputData&& inputData);

    /// Receives tracking parameters
    void ReceiveParameters(Parameters<fvec>&& parameters);

    /// Gets pointer to input data object for external access
    const InputData& GetInputData() const { return fInputData; }

    int PackIndex(const int& a, const int& b, const int& c);

    int UnPackIndex(const int& i, int& a, int& b, int& c);

    /// \brief Sets a default particle mass for the track fit
    ///
    /// The function is used during the reconstruction in order to estimate the multiple scattering and energy loss
    /// \param mass Default particle mass
    void SetDefaultParticleMass(float mass) { fDefaultMass = mass; }

    /// Gets default particle mass
    /// \return particle mass
    float GetDefaultParticleMass() const { return fDefaultMass; }

    /// Gets default particle mass squared
    /// \return particle mass squared
    float GetDefaultParticleMass2() const { return fDefaultMass * fDefaultMass; }


    /*********************************************************************************************/ /**
   *                             ------  FUNCTIONAL PART ------
   ************************************************************************************************/


    void Init(const TrackingMode mode);

    void Finish();

    void PrintHits();

    /// \brief Gets monitor data
    const TrackingMonitorData& GetMonitorData() const { return fMonitorData; }

    /// \brief Sets monitor data
    void SetMonitorData(const TrackingMonitorData& monitorData) { fMonitorData = monitorData; }

   public:
    /// Gets number of stations before the pipe (MVD stations in CBM)
    // int GetNstationsBeforePipe() const { return fNstationsBeforePipe; }

    /// Gets number of stations situated in field region (MVD + STS in CBM)
    int GetNfieldStations() const { return fNfieldStations; }

    /// Get mc track ID for a hit (debug tool)
    int GetMcTrackIdForCaHit(int iHit) const;

    /// Get mc track ID for a hit (debug tool)
    int GetMcTrackIdForWindowHit(int iGridHit) const;

    // const CbmL1MCTrack* GetMcTrackForWindowHit(int iHit) const;

    void SetNofThtreads(int nThreads)
    {
      fNofThreads = nThreads;
      LOG(info) << "ca::Framework: number of threads is set to " << fNofThreads;
    }

    int GetNofThreads() const { return fNofThreads; }


   private:
    int fNstationsBeforePipe{0};                    ///< number of stations before pipe (MVD stations in CBM)
    int fNfieldStations{0};                         ///< number of stations in the field region
    float fDefaultMass{constants::phys::MuonMass};  ///< mass of the propagated particle [GeV/c2]


    // ***************************
    // ** Member variables list **
    // ***************************

    Parameters<fvec> fParameters;  ///< Object of Framework parameters class
    InputData fInputData;    ///< Tracking input data

    Vector<unsigned char> fvHitKeyFlags{
      "Framework::fvHitKeyFlags"};  ///< List of key flags: has been this hit or cluster already used

    TrackingMonitorData fMonitorData{};  ///< Tracking monitor data (statistics per call)

    int fNofThreads = 1;  ///< Number of threads to execute the track-finder

   public:
    Vector<CaHitTimeInfo> fHitTimeInfo;


    /// ----- Data used during track finding -----
    ///


    /// ----------

    double fCaRecoTime{0.};  // time of the track finder + fitter

    Vector<Track> fRecoTracks{"Framework::fRecoTracks"};       ///< reconstructed tracks
    Vector<ca::HitIndex_t> fRecoHits{"Framework::fRecoHits"};  ///< packed hits of reconstructed tracks


    // WARN: Potential race conditions ->
    Vector<int> fHitKeyToTrack{"Framework::fHitKeyToTrack"};  // strip to track pointers


    fvec EventTime{0.f};
    fvec Err{0.f};

    /// --- data used during finding iterations
    unsigned int fCurrentIterationIndex{0};         ///< index of the corrent iteration, needed for debug output

   public:
    std::vector<ca::WindowData> fvWData;                     ///< Intrnal data processed in a time-window
    std::vector<ca::TrackFinderWindow> fvTrackFinderWindow;  ///< Track finder algorithm for the time window
    ca::TrackFinder fTrackFinder{*this};                     ///< Track finder steer class for the entire time slice
    TrackingMode fTrackingMode{kSts};

   private:
    /// =================================  DATA PART  =================================

    /// ----- Different parameters of CATrackFinder -----

    bool fIsTargetField{false};  ///< is the magnetic field present at the target


  } _fvecalignment;


  // ********************************************
  // ** Inline member functions implementation **
  // ********************************************

  // ---------------------------------------------------------------------------------------------------------------------
  //
  [[gnu::always_inline]] inline unsigned int Framework::PackTripletId(unsigned int iStation, unsigned int iTriplet)
  {
#ifndef FAST_CODE
    assert(iStation < constants::size::MaxNstations);
    assert(iTriplet < constants::size::MaxNtriplets);
#endif
    constexpr unsigned int kMoveStation = constants::size::TripletBits;
    return (iStation << kMoveStation) + iTriplet;
  }

  // ---------------------------------------------------------------------------------------------------------------------
  //
  [[gnu::always_inline]] inline unsigned int Framework::TripletId2Station(unsigned int id)
  {
    constexpr unsigned int kMoveStation = constants::size::TripletBits;
    return id >> kMoveStation;
  }

  // ---------------------------------------------------------------------------------------------------------------------
  //
  [[gnu::always_inline]] inline unsigned int Framework::TripletId2Triplet(unsigned int id)
  {
    constexpr unsigned int kTripletMask = (1u << constants::size::TripletBits) - 1u;
    return id & kTripletMask;
  }

}  // namespace cbm::algo::ca
