/* Copyright (C) 2009-2021 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Valentina Akishina, Ivan Kisel, Sergey Gorbunov [committer], Igor Kulakov, Maksym Zyzak */

/*
 *=====================================================
 *
 *  CBM Level 1 4D Reconstruction
 *
 *  Authors: V.Akishina, I.Kisel,  S.Gorbunov, I.Kulakov, M.Zyzak
 *  Documentation: V.Akishina
 *
 *  e-mail : v.akishina@gsi.de
 *
 *=====================================================
 *
 *  Finds tracks using the Cellular Automaton algorithm
 *
 */


#include "CaTrackFinderWindow.h"

#include "AlgoFairloggerCompat.h"
#include "CaBranch.h"
#include "CaFramework.h"
#include "CaGrid.h"
#include "CaGridEntry.h"
#include "CaTrack.h"
#include "CaTrackParam.h"
#include "CaTripletConstructor.h"

#include <algorithm>
#include <cstdio>
#include <fstream>
#include <iostream>
#include <list>
#include <map>
#include <sstream>

// #include "CaToolsDebugger.h"

using Track = cbm::algo::ca::Track;

using namespace cbm::algo::ca;

// ---------------------------------------------------------------------------------------------------------------------
//
TrackFinderWindow::TrackFinderWindow(ca::Framework& algo, ca::WindowData& wData)
  : frAlgo(algo)
  , frWData(wData)
  , fTrackExtender(frAlgo, frWData)
  , fCloneMerger(frAlgo)
  , fTrackFitter(frAlgo, frWData)
{
  for (unsigned int i = 0; i < constants::size::MaxNstations; ++i) {
    fvTriplets[i].SetName(std::stringstream() << "TrackFinderWindow::fTriplets[" << i << "]");
  }
}


// ---------------------------------------------------------------------------------------------------------------------
//
TrackFinderWindow::~TrackFinderWindow() {}

// ---------------------------------------------------------------------------------------------------------------------
//


bool TrackFinderWindow::checkTripletMatch(const ca::Triplet& l, const ca::Triplet& r, fscal& dchi2) const
{
  dchi2 = 1.e20;

  if (r.GetMHit() != l.GetRHit()) return false;
  if (r.GetLHit() != l.GetMHit()) return false;

  if (r.GetMSta() != l.GetRSta()) return false;
  if (r.GetLSta() != l.GetMSta()) return false;


  const float tripletLinkChi2 = frWData.CurrentIteration()->GetTripletLinkChi2();
  if (r.IsMomentumFitted()) {
    assert(l.IsMomentumFitted());

    fscal dqp = l.GetQp() - r.GetQp();
    fscal Cqp = l.GetCqp() + r.GetCqp();

    if (!std::isfinite(dqp)) return false;
    if (!std::isfinite(Cqp)) return false;

    if (dqp * dqp > tripletLinkChi2 * Cqp) {
      return false;  // bad neighbour // CHECKME why do we need recheck it?? (it really change result)
    }
    dchi2 = dqp * dqp / Cqp;
  }
  else {
    fscal dtx = l.GetTx() - r.GetTx();
    fscal Ctx = l.GetCtx() + r.GetCtx();

    fscal dty = l.GetTy() - r.GetTy();
    fscal Cty = l.GetCty() + r.GetCty();

    // it shouldn't happen, but happens sometimes

    if (!std::isfinite(dtx)) return false;
    if (!std::isfinite(dty)) return false;
    if (!std::isfinite(Ctx)) return false;
    if (!std::isfinite(Cty)) return false;

    if (dty * dty > tripletLinkChi2 * Cty) return false;
    if (dtx * dtx > tripletLinkChi2 * Ctx) return false;

    //dchi2 = 0.5f * (dtx * dtx / Ctx + dty * dty / Cty);
    dchi2 = 0.;
  }

  if (!std::isfinite(dchi2)) return false;

  return true;
}


// **************************************************************************************************
// *                                                                                                *
// *                            ------ CATrackFinder procedure ------                               *
// *                                                                                                *
// **************************************************************************************************

// ---------------------------------------------------------------------------------------------------------------------
//
void TrackFinderWindow::ReadWindowData()
{
  int nHits = 0;
  for (int iS = 0; iS < frAlgo.GetParameters().GetNstationsActive(); iS++) {
    frWData.HitStartIndexOnStation(iS) = nHits;
    frWData.NofHitsOnStation(iS)       = frWData.TsHitIndices(iS).size();
    nHits += frWData.NofHitsOnStation(iS);
  }
  frWData.HitStartIndexOnStation(frAlgo.GetParameters().GetNstationsActive()) = nHits;
  frWData.ResetHitData(nHits);

  for (int iS = 0; iS < frAlgo.GetParameters().GetNstationsActive(); iS++) {
    int iFstHit = frWData.HitStartIndexOnStation(iS);
    for (ca::HitIndex_t ih = 0; ih < frWData.TsHitIndices(iS).size(); ++ih) {
      ca::Hit h = frAlgo.fInputData.GetHit(frWData.TsHitIndex(iS, ih));
      h.SetId(frWData.TsHitIndex(iS, ih));
      frWData.Hit(iFstHit + ih) = h;
    }
  }

  if constexpr (fDebug) {
    LOG(info) << "===== Sliding Window hits: ";
    for (int i = 0; i < nHits; ++i) {
      LOG(info) << "    " << frWData.Hit(i).ToString();
    }
    LOG(info) << "===== ";
  }

  fvHitFirstTriplet.reset(nHits, 0);
  fvHitNofTriplets.reset(nHits, 0);


  frWData.RecoTracks().clear();
  frWData.RecoTracks().reserve(2 * nHits / frAlgo.GetParameters().GetNstationsActive());

  frWData.RecoHitIndices().clear();
  frWData.RecoHitIndices().reserve(2 * nHits);

  fvTrackCandidates.clear();
  fvTrackCandidates.reserve(nHits / 10);
  for (unsigned int iS = 0; iS < constants::size::MaxNstations; iS++) {
    int nHitsStation = frWData.TsHitIndices(iS).size();
    fvTriplets[iS].clear();
    fvTriplets[iS].reserve(2 * nHitsStation);
  }
}

void TrackFinderWindow::CaTrackFinderSlice()
{
  /********************************/ /**
   * CATrackFinder routine setting
   ***********************************/

  ReadWindowData();

  for (int iS = 0; iS < frAlgo.GetParameters().GetNstationsActive(); ++iS) {

    bool timeUninitialised = 1;
    fscal lasttime         = 0;
    fscal starttime        = 0;


    fscal gridMinX = -0.1;
    fscal gridMaxX = 0.1;
    fscal gridMinY = -0.1;
    fscal gridMaxY = 0.1;

    for (ca::HitIndex_t ih = 0; ih < frWData.TsHitIndices(iS).size(); ++ih) {
      const ca::Hit& h = frAlgo.fInputData.GetHit(frWData.TsHitIndex(iS, ih));

      if (h.X() < gridMinX) {
        gridMinX = h.X();
      }
      if (h.X() > gridMaxX) {
        gridMaxX = h.X();
      }
      if (h.Y() < gridMinY) {
        gridMinY = h.Y();
      }
      if (h.Y() > gridMaxY) {
        gridMaxY = h.Y();
      }

      const fscal time = h.T();
      assert(std::isfinite(time));
      if (timeUninitialised || lasttime < time) {
        lasttime = time;
      }
      if (timeUninitialised || starttime > time) {
        starttime = time;
      }
      timeUninitialised = false;
    }

    // TODO: changing the grid also changes the result. Investigate why it happens.
    // TODO: find the optimal grid size

    int nSliceHits = frWData.TsHitIndices(iS).size();
    fscal sizeY    = gridMaxY - gridMinY;
    fscal sizeX    = gridMaxX - gridMinX;
    int nBins2D    = 1 + nSliceHits;
    // TODO: SG: the coefficients should be removed
    fscal yStep = 0.3 * sizeY / sqrt(nBins2D);
    fscal xStep = 0.8 * sizeX / sqrt(nBins2D);

    fscal scale = frAlgo.GetParameters().GetStation(iS).GetZ<float>() - frAlgo.GetParameters().GetTargetPositionZ()[0];

    if (yStep > 0.3 * scale) yStep = 0.3 * scale;
    if (yStep < 0.01 * scale) yStep = 0.01 * scale;
    if (xStep > 0.3 * scale) xStep = 0.3 * scale;
    if (xStep < 0.01 * scale) xStep = 0.01 * scale;

    auto& grid = frWData.Grid(iS);
    grid.BuildBins(gridMinX, gridMaxX, gridMinY, gridMaxY, xStep, yStep);
    /* clang-format off */
    grid.StoreHits(frWData.Hits(), 
                   frWData.HitStartIndexOnStation(iS), 
                   frWData.NofHitsOnStation(iS),
                   frAlgo.fvHitKeyFlags);
    /* clang-format on */
  }


  // ---- Loop over Track Finder iterations ----------------------------------------------------------------//

  for (frAlgo.fCurrentIterationIndex = 0;
       frAlgo.fCurrentIterationIndex < frAlgo.GetParameters().GetCAIterations().size();
       ++frAlgo.fCurrentIterationIndex) {

    // current iteration of the CA tracker
    const auto& caIteration   = frAlgo.GetParameters().GetCAIterations()[frAlgo.fCurrentIterationIndex];
    frWData.SetCurrentIteration(&caIteration);

    if (frAlgo.fCurrentIterationIndex > 0) {
      for (int ista = 0; ista < frAlgo.GetParameters().GetNstationsActive(); ++ista) {
        frWData.Grid(ista).RemoveUsedHits(frWData.Hits(), frAlgo.fvHitKeyFlags);
      }
    }

    int nHits = frWData.Hits().size();
    // --> frAlgo.fIsWindowHitSuppressed.reset(frAlgo.fWindowHits.size(), 0);
    frWData.ResetHitSuppressionFlags();  // TODO: ??? No effect?
    for (int j = 0; j < frAlgo.GetParameters().GetNstationsActive(); j++) {
      fvTriplets[j].clear();
    }

    fvHitFirstTriplet.reset(nHits, 0);
    fvHitNofTriplets.reset(nHits, 0);

    {
      // #pragma omp  task
      {
        // --- SET PARAMETERS FOR THE ITERATION ---
        // define the target
        fscal SigmaTargetX = caIteration.GetTargetPosSigmaX();
        fscal SigmaTargetY = caIteration.GetTargetPosSigmaY();  // target constraint [cm]

        // Select magnetic field. For primary tracks - fVtxFieldValue, for secondary tracks - st.fieldSlice
        if (caIteration.GetPrimaryFlag()) {
          frWData.TargB() = frAlgo.GetParameters().GetVertexFieldValue();
        }
        else {
          frWData.TargB() = frAlgo.GetParameters().GetStation(0).fieldSlice.GetFieldValue(0, 0);
        }  // NOTE: calculates field frAlgo.fTargB in the center of 0th station

        frAlgo.fIsTargetField = (fabs(frWData.TargB().y[0]) > 0.001);

        frWData.TargetMeasurement().SetX(frAlgo.GetParameters().GetTargetPositionX());
        frWData.TargetMeasurement().SetY(frAlgo.GetParameters().GetTargetPositionY());
        frWData.TargetMeasurement().SetDx2(SigmaTargetX * SigmaTargetX);
        frWData.TargetMeasurement().SetDxy(0);
        frWData.TargetMeasurement().SetDy2(SigmaTargetY * SigmaTargetY);
        frWData.TargetMeasurement().SetNdfX(1);
        frWData.TargetMeasurement().SetNdfY(1);
      }
    }


    ///   stage for triplets creation

    frAlgo.fMonitorData.StartTimer(ETimer::TripletConstruction);

    ca::TripletConstructor constructor(frAlgo, frWData);

    // indices of the two neighbouring station, taking into account allowed gaps
    std::vector<std::pair<int, int>> staPattern;
    for (int gap = 0; gap <= frWData.CurrentIteration()->GetMaxStationGap(); gap++) {
      for (int i = 0; i <= gap; i++) {
        staPattern.push_back(std::make_pair(1 + i, 2 + gap));
      }
    }

    const int iStFirst = frWData.CurrentIteration()->GetFirstStationIndex();
    for (int istal = frAlgo.GetParameters().GetNstationsActive() - 2; istal >= iStFirst; istal--) {
      // start with downstream chambers
      const auto& grid     = frWData.Grid(istal);
      Tindex nGridEntriesL = grid.GetEntries().size();
      for (Tindex iel = 0; iel < nGridEntriesL; ++iel) {
        ca::HitIndex_t ihitl = grid.GetEntries()[iel].GetObjectId();
        auto oldSize         = fvTriplets[istal].size();
        for (auto& pattern : staPattern) {
          constructor.CreateTripletsForHit(istal, istal + pattern.first, istal + pattern.second, ihitl);
          fvTriplets[istal].insert(fvTriplets[istal].end(), constructor.GetTriplets().begin(),
                                   constructor.GetTriplets().end());
        }
        fvHitFirstTriplet[ihitl] = frAlgo.PackTripletId(istal, oldSize);
        fvHitNofTriplets[ihitl]  = fvTriplets[istal].size() - oldSize;
      }
    }  // istal

    frAlgo.fMonitorData.StopTimer(ETimer::TripletConstruction);

    // search for neighbouring triplets
    frAlgo.fMonitorData.StartTimer(ETimer::NeighboringTripletSearch);

    for (int istal = frAlgo.GetParameters().GetNstationsActive() - 2; istal >= iStFirst; istal--) {
      // start with downstream chambers
      for (unsigned int it = 0; it < fvTriplets[istal].size(); ++it) {

        ca::Triplet& tr = fvTriplets[istal][it];
        tr.SetLevel(0);
        tr.SetFNeighbour(0);
        tr.SetNNeighbours(0);

        unsigned int nNeighbours = fvHitNofTriplets[tr.GetMHit()];

        unsigned int neighLocation = fvHitFirstTriplet[tr.GetMHit()];
        unsigned int neighStation  = frAlgo.TripletId2Station(neighLocation);
        unsigned int neighTriplet  = frAlgo.TripletId2Triplet(neighLocation);

        if (nNeighbours > 0) {
          assert((int) neighStation >= istal + 1
                 && (int) neighStation <= istal + 1 + frWData.CurrentIteration()->GetMaxStationGap());
        }

        unsigned char level = 0;

        for (unsigned int iN = 0; iN < nNeighbours; ++iN, ++neighTriplet, ++neighLocation) {

          ca::Triplet& neighbour = fvTriplets[neighStation][neighTriplet];

          fscal dchi2 = 0.;
          if (!checkTripletMatch(tr, neighbour, dchi2)) continue;

          if (tr.GetFNeighbour() == 0) {
            tr.SetFNeighbour(neighLocation);
          }

          tr.SetNNeighbours(neighLocation - tr.GetFNeighbour() + 1);

          const unsigned char neighbourLevel = neighbour.GetLevel();

          if (level <= neighbourLevel) {
            level = neighbourLevel + 1;
          }
        }
        tr.SetLevel(level);
      }  // neighbour search

      frAlgo.fMonitorData.IncrementCounter(ECounter::Triplet, fvTriplets[istal].size());

    }  // istal
    frAlgo.fMonitorData.StopTimer(ETimer::NeighboringTripletSearch);


    ///====================================================================
    ///=                                                                  =
    ///=        Collect track candidates. CREATE TRACKS                   =
    ///=                                                                  =
    ///====================================================================

    // min level to start triplet. So min track length = min_level+3.
    int min_level = std::min(caIteration.GetMinNhits(), caIteration.GetMinNhitsStation0()) - 3;

    if (frWData.CurrentIteration()->GetTrackFromTripletsFlag()) {
      min_level = 0;
    }

    ca::Branch curr_tr;
    ca::Branch new_tr[constants::size::MaxNstations];
    ca::Branch best_tr;

    int ndf = 1;


    // collect consequtive: the longest tracks, shorter, more shorter and so on
    for (int firstTripletLevel = frAlgo.GetParameters().GetNstationsActive() - 3; firstTripletLevel >= min_level;
         firstTripletLevel--) {
      // choose length in triplets number - firstTripletLevel - the maximum possible triplet level among all triplets in the searched candidate

      //  how many levels to check
      int nlevel = (frAlgo.GetParameters().GetNstationsActive() - 2) - firstTripletLevel + 1;

      const unsigned char min_best_l =
        (firstTripletLevel > min_level) ? firstTripletLevel + 2 : min_level + 3;  // loose maximum


      fvTrackCandidates.clear();

      for (const auto& h : frWData.Hits()) {
        frAlgo.fHitKeyToTrack[h.FrontKey()] = -1;
        frAlgo.fHitKeyToTrack[h.BackKey()]  = -1;
      }

      //== Loop over triplets with the required level, find and store track candidates

      for (int istaF = iStFirst; istaF <= frAlgo.GetParameters().GetNstationsActive() - 3 - firstTripletLevel;
           ++istaF) {

        if (--nlevel == 0) break;  //TODO: SG: this is not needed

        for (Tindex itrip = 0; itrip < (Tindex) fvTriplets[istaF].size(); ++itrip) {

          ca::Triplet& first_trip = (fvTriplets[istaF][itrip]);
          const auto& fstTripLHit = frWData.Hit(first_trip.GetLHit());
          if (frAlgo.fvHitKeyFlags[fstTripLHit.FrontKey()] || frAlgo.fvHitKeyFlags[fstTripLHit.BackKey()]) {
            continue;
          }

          //  skip track candidates that are too short

          int minNhits = frWData.CurrentIteration()->GetMinNhits();

          if (fstTripLHit.Station() == 0) {
            minNhits = frWData.CurrentIteration()->GetMinNhitsStation0();
          }
          if (frWData.CurrentIteration()->GetTrackFromTripletsFlag()) {
            minNhits = 0;
          }

          if (3 + first_trip.GetLevel() < minNhits) {
            continue;
          }

          // Collect triplets, which can start a track with length equal to firstTipletLevel + 3. This cut suppresses
          // ghost tracks, but does not affect the efficiency
          if (first_trip.GetLevel() < firstTripletLevel) {
            continue;
          }

          curr_tr.SetChi2(0.);
          curr_tr.SetStation(0);
          curr_tr.ResetHits();
          curr_tr.AddHit(fstTripLHit.Id());
          curr_tr.SetChi2(first_trip.GetChi2());

          best_tr = curr_tr;

          CAFindTrack(istaF, best_tr, &first_trip, curr_tr, min_best_l,
                      new_tr);  /// reqursive func to build a tree of possible track-candidates and choose the best

          if (best_tr.NofHits() < firstTripletLevel + 2) continue;  // loose maximum one hit

          if (best_tr.NofHits() < min_level + 3) continue;  // should find all hits for min_level

          if (best_tr.NofHits() < minNhits) {
            continue;
          }

          ndf = best_tr.NofHits() * 2 - 5;

          // TODO: automatize the NDF calculation

          if (ca::Framework::TrackingMode::kGlobal == frAlgo.fTrackingMode
              || ca::Framework::TrackingMode::kMcbm == frAlgo.fTrackingMode) {
            ndf = best_tr.NofHits() * 2 - 4;
          }

          best_tr.SetChi2(best_tr.Chi2() / ndf);
          if (frAlgo.GetParameters().GetGhostSuppression()) {
            if (3 == best_tr.NofHits()) {
              if (!frWData.CurrentIteration()->GetPrimaryFlag() && (istaF != 0))
                continue;  // too /*short*/ non-MAPS track
              if (frWData.CurrentIteration()->GetPrimaryFlag() && (best_tr.Chi2() > 5.0)) continue;
            }
          }
          fvTrackCandidates.push_back(best_tr);
          ca::Branch& tr = fvTrackCandidates.back();
          tr.SetStation(istaF);
          tr.SetId(fvTrackCandidates.size() - 1);
          tr.SetAlive(true);
          if constexpr (fDebug) {
            std::stringstream s;
            s << "iter " << frAlgo.fCurrentIterationIndex << ", track candidate " << fvTrackCandidates.size() - 1
              << " found, L = " << best_tr.NofHits() << " chi2= " << best_tr.Chi2() << " hits: ";
            for (auto hitId : tr.Hits()) {
              s << hitId << " (mc " << frAlgo.GetMcTrackIdForCaHit(hitId) << ") ";
            }
            LOG(info) << s.str();
          }
        }  // itrip
      }    // istaF

      for (ca::Branch& tr : fvTrackCandidates) {
        tr.SetAlive(false);
      }

      bool repeatCompetition = true;
      int iComp              = 0;
      for (iComp = 0; (iComp < 100) && repeatCompetition; ++iComp) {

        // == Loop over track candidates and mark their strips

        repeatCompetition = false;

        for (ca::Branch& tr : fvTrackCandidates) {

          if (tr.IsAlive()) {
            continue;
          }
          for (auto& hitId : tr.Hits()) {
            const ca::Hit& h = frAlgo.fInputData.GetHit(hitId);
            bool isAlive     = true;
            {  // front  strip
              auto& stripF = (frAlgo.fHitKeyToTrack)[h.FrontKey()];
              if ((stripF >= 0) && (stripF != tr.Id())) {  // strip is used by other candidate
                const auto& other = fvTrackCandidates[stripF];
                if (!other.IsAlive() && tr.IsBetterThan(other)) {
                  stripF = tr.Id();
                }
                else {
                  isAlive = false;
                }
              }
              else {
                stripF = tr.Id();
              }
              if (!isAlive) {
                break;
              }
            }

            {  // back strip
              auto& stripB = (frAlgo.fHitKeyToTrack)[h.BackKey()];
              if ((stripB >= 0) && (stripB != tr.Id())) {  // strip is used by other candidate
                const auto& other = fvTrackCandidates[stripB];
                if (!other.IsAlive() && tr.IsBetterThan(other)) {
                  stripB = tr.Id();
                }
                else {
                  isAlive = false;
                }
              }
              else {
                stripB = tr.Id();
              }
              if (!isAlive) {
                break;
              }
            }
          }  // loop over hits
        }    // itrack

        // == Check if some suppressed candidates are still alive

        for (ca::Branch& tr : fvTrackCandidates) {
          if (tr.IsAlive()) {
            continue;
          }

          tr.SetAlive(true);
          for (int iHit = 0; tr.IsAlive() && (iHit < (int) tr.Hits().size()); ++iHit) {
            const ca::Hit& h = frAlgo.fInputData.GetHit(tr.Hits()[iHit]);
            tr.SetAlive(tr.IsAlive() && ((frAlgo.fHitKeyToTrack)[h.FrontKey()] == tr.Id())
                        && ((frAlgo.fHitKeyToTrack)[h.BackKey()] == tr.Id()));
          }

          if (!tr.IsAlive()) {  // release strips
            for (auto hitId : tr.Hits()) {
              const ca::Hit& h = frAlgo.fInputData.GetHit(hitId);
              if (frAlgo.fHitKeyToTrack[h.FrontKey()] == tr.Id()) {
                frAlgo.fHitKeyToTrack[h.FrontKey()] = -1;
              }
              if (frAlgo.fHitKeyToTrack[h.BackKey()] == tr.Id()) {
                frAlgo.fHitKeyToTrack[h.BackKey()] = -1;
              }
            }
          }
          else {
            repeatCompetition = true;
          }
        }  // itrack

      }  // competitions

      // LOG(info) << " N Competitions: " << iComp ;

      // ==

      for (Tindex iCandidate = 0; iCandidate < (Tindex) fvTrackCandidates.size(); ++iCandidate) {
        ca::Branch& tr = fvTrackCandidates[iCandidate];

        if constexpr (fDebug) {
          LOG(info) << "iter " << frAlgo.fCurrentIterationIndex << ", track candidate " << iCandidate
                    << ": alive = " << tr.IsAlive();
        }

        if (!tr.IsAlive()) continue;

        if (frWData.CurrentIteration()->GetExtendTracksFlag()) {
          if (tr.NofHits() < frAlgo.GetParameters().GetNstationsActive()) {
            fTrackExtender.ExtendBranch(tr);
          }
        }

        for (auto iHit : tr.Hits()) {
          const ca::Hit& hit = frAlgo.fInputData.GetHit(iHit);

          /// used strips are marked

          frAlgo.fvHitKeyFlags[hit.FrontKey()] = 1;
          frAlgo.fvHitKeyFlags[hit.BackKey()]  = 1;

          frWData.RecoHitIndices().push_back(iHit);
        }
        Track t;
        t.fNofHits = tr.NofHits();
        frWData.RecoTracks().push_back(t);
        if (0) {  // SG debug
          std::stringstream s;
          s << "store track " << iCandidate << " chi2= " << tr.Chi2() << "\n";
          s << " hits: ";
          for (unsigned int ih = 0; ih < tr.Hits().size(); ih++) {
            s << frAlgo.GetMcTrackIdForCaHit(tr.Hits()[ih]) << " ";
          }
          LOG(info) << s.str();
        }

      }  // tracks

    }  // firstTripletLevel


    // suppress strips of suppressed hits
    for (unsigned int ih = 0; ih < frWData.Hits().size(); ih++) {
      if (frWData.IsHitSuppressed(ih)) {
        const ca::Hit& hit                   = frWData.Hit(ih);
        frAlgo.fvHitKeyFlags[hit.FrontKey()] = 1;
        frAlgo.fvHitKeyFlags[hit.BackKey()]  = 1;
      }
    }
  }  // ---- Loop over Track Finder iterations: END -----------------------------------------------------------//

  // Fit tracks
  frAlgo.fMonitorData.StartTimer(ETimer::TrackFitter);
  fTrackFitter.FitCaTracks();
  frAlgo.fMonitorData.StopTimer(ETimer::TrackFitter);

  // Merge clones
  fCloneMerger.Exec(frWData.RecoTracks(), frWData.RecoHitIndices());

  frAlgo.fMonitorData.StartTimer(ETimer::TrackFitter);
  fTrackFitter.FitCaTracks();
  frAlgo.fMonitorData.StopTimer(ETimer::TrackFitter);
}


/** *************************************************************
     *                                                              *
     *     The routine performs recursive search for tracks         *
     *                                                              *
     *     I. Kisel                                    06.03.05     *
     *     I.Kulakov                                    2012        *
     *                                                              *
     ****************************************************************/

void TrackFinderWindow::CAFindTrack(int ista, ca::Branch& best_tr, const ca::Triplet* curr_trip, ca::Branch& curr_tr,
                                    unsigned char min_best_l, ca::Branch* new_tr)
/// recursive search for tracks
/// input: @ista - station index, @&best_tr - best track for the privious call
/// output: @&NCalls - number of function calls
{

  if (curr_trip->GetLevel() == 0)  // the end of the track -> check and store
  {
    // -- finish with current track
    // add rest of hits
    const auto& hitM = frWData.Hit(curr_trip->GetMHit());
    const auto& hitR = frWData.Hit(curr_trip->GetRHit());

    if (!(frAlgo.fvHitKeyFlags[hitM.FrontKey()] || frAlgo.fvHitKeyFlags[hitM.BackKey()])) {
      curr_tr.AddHit(hitM.Id());
    }

    if (!(frAlgo.fvHitKeyFlags[hitR.FrontKey()] || frAlgo.fvHitKeyFlags[hitR.BackKey()])) {
      curr_tr.AddHit(hitR.Id());
    }

    //if( curr_tr.NofHits() < min_best_l - 1 ) return; // suppouse that only one hit can be added by extender
    // TODO: SZh 21.08.2023: Replace hard-coded value with a parameter
    int ndf = curr_tr.NofHits() * 2 - 5;

    if (ca::Framework::TrackingMode::kGlobal == frAlgo.fTrackingMode
        || ca::Framework::TrackingMode::kMcbm == frAlgo.fTrackingMode) {
      ndf = curr_tr.NofHits() * 2 - 4;
    }


    if (curr_tr.Chi2() > frWData.CurrentIteration()->GetTrackChi2Cut() * ndf) {
      return;
    }


    // -- select the best
    if ((curr_tr.NofHits() > best_tr.NofHits())
        || ((curr_tr.NofHits() == best_tr.NofHits()) && (curr_tr.Chi2() < best_tr.Chi2()))) {
      best_tr = curr_tr;
    }
    return;
  }
  else  //MEANS level ! = 0
  {
    int N_neighbour = (curr_trip->GetNNeighbours());

    for (Tindex in = 0; in < N_neighbour; in++) {

      unsigned int ID = curr_trip->GetFNeighbour() + in;

      unsigned int Station = frAlgo.TripletId2Station(ID);
      unsigned int Triplet = frAlgo.TripletId2Triplet(ID);

      const ca::Triplet& new_trip = fvTriplets[Station][Triplet];

      fscal dchi2 = 0.;
      if (!checkTripletMatch(*curr_trip, new_trip, dchi2)) continue;

      const auto& hitL = frWData.Hit(new_trip.GetLHit());                                   // left hit of new triplet
      if (frAlgo.fvHitKeyFlags[hitL.FrontKey()] || frAlgo.fvHitKeyFlags[hitL.BackKey()]) {  //hits are used
        //  no used hits allowed -> compare and store track
        if ((curr_tr.NofHits() > best_tr.NofHits())
            || ((curr_tr.NofHits() == best_tr.NofHits()) && (curr_tr.Chi2() < best_tr.Chi2()))) {
          best_tr = curr_tr;
        }
      }
      else {  // hit is not used: add the left hit from the new triplet to the current track

        unsigned char new_L = curr_tr.NofHits() + 1;
        fscal new_chi2      = curr_tr.Chi2() + dchi2;


        if (0) {  //SGtrd2d debug!!
          int mc01 = frAlgo.GetMcTrackIdForWindowHit(curr_trip->GetLHit());
          int mc02 = frAlgo.GetMcTrackIdForWindowHit(curr_trip->GetMHit());
          int mc03 = frAlgo.GetMcTrackIdForWindowHit(curr_trip->GetRHit());
          int mc11 = frAlgo.GetMcTrackIdForWindowHit(new_trip.GetLHit());
          int mc12 = frAlgo.GetMcTrackIdForWindowHit(new_trip.GetMHit());
          int mc13 = frAlgo.GetMcTrackIdForWindowHit(new_trip.GetRHit());
          if ((mc01 == mc02) && (mc02 == mc03)) {
            LOG(info) << " sta " << ista << " mc0 " << mc01 << " " << mc02 << " " << mc03 << " mc1 " << mc11 << " "
                      << mc12 << " " << mc13 << " chi2 " << curr_tr.Chi2() / (2 * (curr_tr.NofHits() + 2) - 4)
                      << " new " << new_chi2 / (2 * (new_L + 2) - 4);
            LOG(info) << "   hits " << curr_trip->GetLHit() << " " << curr_trip->GetMHit() << " "
                      << curr_trip->GetRHit() << " " << new_trip.GetLHit();
          }
        }

        int ndf = 2 * (new_L + 2) - 5;

        if (ca::Framework::TrackingMode::kGlobal == frAlgo.fTrackingMode) {  //SGtrd2d!!!
          ndf = 2 * (new_L + 2) - 4;
        }
        else if (ca::Framework::TrackingMode::kMcbm == frAlgo.fTrackingMode) {
          ndf = 2 * (new_L + 2) - 4;
        }
        else {
          ndf = new_L;  // TODO: SG:  2 * (new_L + 2) - 5
        }

        if (new_chi2 > frWData.CurrentIteration()->GetTrackChi2Cut() * ndf) continue;

        // add new hit
        new_tr[ista] = curr_tr;
        new_tr[ista].AddHit(hitL.Id());

        const int new_ista = ista + new_trip.GetMSta() - new_trip.GetLSta();
        new_tr[ista].SetChi2(new_chi2);

        CAFindTrack(new_ista, best_tr, &new_trip, new_tr[ista], min_best_l, new_tr);
      }  // add triplet to track
    }    // for neighbours
  }      // level = 0
}
