/* Copyright (C) 2024 FIAS Frankfurt Institute for Advanced Studies, Frankfurt / Main
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Felix Weiglhofer [committer], Dominik Smith */

#include "Unpack.h"

#include "log.hpp"

using namespace cbm::algo::sts;
using fles::Subsystem;

Unpack::Unpack(const ReadoutConfigLegacy& readout) : fReadout(readout)
{
  uint32_t numChansPerAsicSts   = 128;  // R/O channels per ASIC for STS
  uint32_t numAsicsPerModuleSts = 16;   // Number of ASICs per module for STS

  constexpr i64 SystemTimeOffset = -970;  // ns?

  auto equipIdsSts = fReadout.GetEquipmentIds();
  for (auto& equip : equipIdsSts) {
    std::unique_ptr<sts::UnpackPar> par(new sts::UnpackPar());
    par->fNumChansPerAsic   = numChansPerAsicSts;
    par->fNumAsicsPerModule = numAsicsPerModuleSts;
    const size_t numElinks  = fReadout.GetNumElinks(equip);
    for (size_t elink = 0; elink < numElinks; elink++) {
      sts::UnpackElinkPar elinkPar;
      auto mapEntry        = fReadout.Map(equip, elink);
      elinkPar.fAddress    = mapEntry.first;   // Module address for this elink
      elinkPar.fAsicNr     = mapEntry.second;  // ASIC number within module
      elinkPar.fTimeOffset = SystemTimeOffset;
      elinkPar.fAdcMinCut  = fReadout.AdcCutMap(equip, elink);
      elinkPar.fAdcOffset  = 1.;
      elinkPar.fAdcGain    = 1.;
      elinkPar.fWalk       = fReadout.WalkMap(elinkPar.fAddress, elinkPar.fAsicNr);
      elinkPar.fChanMask   = fReadout.MaskMap(equip, elink);
      // TODO: Add parameters for time and ADC calibration
      par->fElinkParams.push_back(elinkPar);
    }
    fAlgos[equip].SetParams(std::move(par));
    L_(debug) << "--- Configured equipment " << equip << " with " << numElinks << " elinks";
  }  //# equipments
}

Unpack::Result_t Unpack::operator()(const fles::Timeslice& ts) const
{
  constexpr int SystemVersion = 0x20;
  return DoUnpack(Subsystem::STS, ts, SystemVersion);
}
