/* Copyright (C) 2024 FIAS Frankfurt Institute for Advanced Studies, Frankfurt / Main
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Felix Weiglhofer [committer], Dominik Smith */
#pragma once

#include "Definitions.h"
#include "PODVector.h"
#include "Timeslice.hpp"
#include "compat/Algorithm.h"
#include "compat/OpenMP.h"

#include <cstdint>
#include <gsl/span>
#include <map>
#include <vector>

#include <xpu/host.h>

namespace cbm::algo
{

  namespace detail
  {

    struct UnpackMonitorBase {
      fles::Subsystem system;       // subsystem
      size_t numMs            = 0;  // number of microslices
      size_t sizeBytesIn      = 0;  // total size of microslice contents
      size_t sizeBytesOut     = 0;  // total size of unpacked digis
      size_t errInvalidSysVer = 0;
      size_t errInvalidEqId   = 0;

      double ExpansionFactor() const { return sizeBytesIn > 0 ? static_cast<double>(sizeBytesOut) / sizeBytesIn : 0.0; }
    };

    /**
     * @brief Collection of MS data to unpack
     *
     * @note Helper struct for CommonUnpacker. Moved out of class body to avoid code duplication from different template specializations.
     */
    struct MSData {
      std::vector<u16> msEqIds;                        // equipment ids of microslices
      std::vector<fles::MicrosliceDescriptor> msDesc;  // microslice descriptors
      std::vector<const u8*> msContent;                // pointer to microslice contents

      UnpackMonitorBase monitor;  // monitoring data

      MSData(const fles::Timeslice& ts, fles::Subsystem system, gsl::span<u16> legalEqIds, u8 sys_ver);
      ~MSData() = default;
    };

  }  // namespace detail

  template<class MSMonitor>
  struct UnpackMonitor : detail::UnpackMonitorBase {
    std::vector<MSMonitor> msMonitor;  // monitoring data per microslice
  };

  template<class Digi, class UnpackAlgo, class MSMonitor>
  class CommonUnpacker {

   protected:
    using Monitor_t = UnpackMonitor<MSMonitor>;
    using Result_t  = std::pair<PODVector<Digi>, Monitor_t>;

    std::map<u16, UnpackAlgo> fAlgos;  //< Equipment id to unpacker map. Filled by child class!

    Result_t DoUnpack(const fles::Subsystem subsystem, const fles::Timeslice& ts, u8 sys_ver) const
    {
      xpu::scoped_timer t_(fles::to_string(subsystem));
      auto legalEqIds = GetEqIds();

      detail::MSData msData{ts, subsystem, gsl::make_span(legalEqIds), sys_ver};
      const size_t numMs = msData.monitor.numMs;

      Result_t out;
      auto& digisOut   = out.first;
      auto& monitorOut = out.second;

      static_cast<detail::UnpackMonitorBase&>(monitorOut) = msData.monitor;

      std::vector<std::vector<Digi>> msDigis(numMs);  // unpacked digis per microslice
      monitorOut.msMonitor.resize(numMs);             // monitoring data per microslice

      xpu::t_add_bytes(msData.monitor.sizeBytesIn);

      xpu::push_timer("Unpack");
      xpu::t_add_bytes(msData.monitor.sizeBytesIn);
      CBM_PARALLEL_FOR(schedule(dynamic))
      for (size_t i = 0; i < numMs; i++) {
        auto result             = fAlgos.at(msData.msEqIds[i])(msData.msContent[i], msData.msDesc[i], ts.start_time());
        msDigis[i]              = std::move(result.first);
        monitorOut.msMonitor[i] = std::move(result.second);
      }
      xpu::pop_timer();

      xpu::push_timer("Resize");
      size_t nDigisTotal = 0;
      for (const auto& digis : msDigis) {
        nDigisTotal += digis.size();
      }
      digisOut.resize(nDigisTotal);
      monitorOut.sizeBytesOut = nDigisTotal * sizeof(Digi);
      xpu::pop_timer();

      xpu::push_timer("Merge");
      xpu::t_add_bytes(monitorOut.sizeBytesOut);
      CBM_PARALLEL_FOR(schedule(dynamic))
      for (size_t i = 0; i < numMs; i++) {
        size_t offset = 0;
        for (size_t x = 0; x < i; x++)
          offset += msDigis[x].size();
        std::copy(msDigis[i].begin(), msDigis[i].end(), digisOut.begin() + offset);
      }
      xpu::pop_timer();

      DoSort(digisOut);

      return out;
    }

   private:
    void DoSort(gsl::span<Digi> digis) const
    {
      Sort(digis.begin(), digis.end(), [](const Digi& a, const Digi& b) { return a.GetTime() < b.GetTime(); });
    }

    std::vector<uint16_t> GetEqIds() const
    {
      std::vector<uint16_t> eqIds;
      eqIds.reserve(fAlgos.size());
      for (const auto& [eqId, algo] : fAlgos) {
        eqIds.push_back(eqId);
      }
      return eqIds;
    }
  };
}  // namespace cbm::algo
